from . import getsta, getfreq, schfiles, gettim, gintent, toggle, infdb, \
    invla, schrep, scndup, getcov, getcor, times, sttant, parameter, schdefs
from ..catalog import ScanCatalog, SetupFileCatalog
from .. import util, key

import schedlib as s

import numpy as np

import sys

maxsta = s.schn1.stanum.shape[0]
maxchan = s.schn2b.freq.shape[0]

record_defaults = {
    "comment":  ["",                    util.noop],
    "crddop":   [1.,                    util.to_bool],
    "crdnodop": [1.,                    util.to_bool],
    "doppler":  [1.,                    util.to_bool],
    "nodop":    [1.,                    util.to_bool],
    "duration": [parameter.unset,       util.noop],
    "freqlist": [[parameter.unset] * 2, util.noop],
    "geoseg":   [0.,                    util.multiply_by(parameter.onesec)],
    "group":    [1,                     util.noop],
    "repeat":   [1,                     util.noop],
    "intents":  [[],                    util.noop],
    "record":   [1.,                    util.to_bool],
    "norecord": [1.,                    util.to_bool],
    "ptvlba":   [1.,                    util.to_bool],
    "noptvlba": [1.,                    util.to_bool],
    "tavlba":   [1.,                    util.to_bool],
    "notavlba": [1.,                    util.to_bool],
    "tsys":     [1.,                    util.to_bool],
    "notsys":   [1.,                    util.to_bool],
    "pn3db":    [1.,                    util.to_bool],
    "nopn3db":  [1.,                    util.to_bool],
    "point":    [-999,                  util.noop],
    "scantag":  ["",                    util.noop],
    "start":    [parameter.unset,       util.noop],
    "stop":     [parameter.unset,       util.noop],
    "tapefile": [None,                  util.noop],
    "vlatsys":  [None,                  util.noop],
    "vlantsys": [None,                  util.noop],
    "nopeak":   [None,                  util.noop],
    "exit":     [None,                  util.noop],
    "lineinit": [None,                  util.noop],
    "srccat":   [None,                  util.noop],
    "peakinit": [None,                  util.noop],
    "stacat":   [None,                  util.noop],
    "satinit":  [None,                  util.noop],
    "setinit":  [None,                  util.noop],
    "pcenters": [None,                  util.noop],
    "coverlet": [None,                  util.noop],
    "nchan":    [None,                  util.noop],
    "autotape": [None,                  util.noop],
    "tape":     [None,                  util.noop],
    "fastfor":  [None,                  util.noop],
    "reverse":  [None,                  util.noop],
    "stations": [[],                    util.foreach(util.upper)],
}

# if an inline catalog is read, these values have to be passed to the next 
# record read
keep_for_next = {
    'start', 'stop', 'repeat', 'group', 'duration', 'dwell', 'nopeak', 
    'point', 'scantag', 'geoseg', 'higroup', 'doppler', 'nodop', 'crddop', 
    'crdnodop', 'record', 'norecord', 'ptvlba', 'noptvlba', 'tavlba', 
    'notavlba', 'tsys', 'notsys', 'vlatsys', 'vlantsys', 'stations', 
    'comment', 'intents', 'scanexps' }

state_defaults = {
    "schedule":  ["",                      util.noop],
    "freqfile":  [schdefs("frequency"),    util.noop],
    "stafile":   [schdefs("stations"),     util.noop],
    "locfile":   [schdefs("location"),     util.noop],
    "msgfile":   [schdefs("messages"),     util.noop],
    "peakfile":  [schdefs("peakcommand"),  util.noop],
    "srcfile":   [schdefs("sources"),      util.noop],
    "srcfile2":  ["NONE",                  util.noop],
    "ephfile":   ["NONE",                  util.noop],
    "dosta":     ["ALL",                   util.upper],
    "day":       [0,                       util.noop],
    "month":     [1,                       util.noop],
    "year":      [0,                       util.noop],
    "dwell":     [[0.] * 3,                util.noop],
    "gap":       [0.,                      util.multiply_by(1/86400.)],
    "prescan":   [0.,                      util.multiply_by(1/86400.)],
    "datapath":  ["IN2DISK",               util.upper],
    "grabto":    ["NONE",                  util.upper],
    "grabtime":  [[parameter.unset] * 2,   util.noop],
    "grabgap":   [0.,                      util.noop],
    "source":    ["DUMMY",                 util.upper],
    "qual":      [0.,                      util.noop],
    "centers":   ["",                      util.upper],
    "tant1":     [0.,                      util.to_bool],
    "tant2":     [0.,                      util.to_bool],
    "caltime":   [120.,                    util.noop],
    "ptslew":    [160.,                    util.noop],
    "pcal":      ["",                      util.upper],
    "focus":     [0.,                      util.noop],
    "rotation":  [0.,                      util.noop],
    "azcolim":   [0.,                      util.noop],
    "elcolim":   [0.,                      util.noop],
    "opmiss":    [0.,                      util.noop],
    "crdline":   ["",                      util.noop],
    "dodown":    [1.,                      util.to_bool],
    "preempt":   ["--",                    util.upper],
    "peak":      [parameter.unset,         util.noop],
    "autopeak":  [1.,                      util.to_bool],
    "pkwatch":   [1.,                      util.to_bool],
    "higroup":   [1.,                      util.noop],
    "setup":     ["DUMMY",                 util.noop],
    "freq":      [[0.],                    util.chain(
        util.foreach(lambda x: round(x, 7)), util.extend_to(maxchan))],
    "bw":        [[0.],                    util.chain(
        util.foreach(lambda x: round(x, 7)), util.extend_to(maxchan))],
    "crdnch":    [0,                       int],
    "crdch1":    [0.,                      util.noop],
    "crdsetch":  [[0.] * 4,                util.noop],
    "crdfreq":   [[0.],                    util.chain(
        util.foreach(lambda x: round(x, 7)), util.extend_to(maxchan))],
    "crdbw":     [[0.],                    util.chain(
        util.foreach(lambda x: round(x, 7)), util.extend_to(maxchan))],
    "dopsrc":    ["",                      util.upper],
    "dopcal":    [1.,                      util.to_bool],
    "linename":  ["",                      util.upper],
    "dopincr":   [[0., 0.],                util.noop],
    "vlamode":   ["ZZ",                    util.upper],
    "vlapeak":   ["OFF",                   util.upper],
    "vlapsrc":   ["",                      util.upper],
    "vlabw":     [None,                    util.noop],
    "vlaband":   [None,                    util.noop],
    "vlainteg":  [0.,                      util.noop],
    "vlaptime":  [10.,                     util.noop],
    "vlatype":   ["VLBI",                  util.noop],
    "vlausern":  [600.,                    util.noop],
    "iatutc":    [0.,                      util.noop],
    "vlarfant":  [10.,                     util.noop],
    "opminel":   [2.,                      util.noop],
    "opminant":  [0.,                      util.noop],
    "opminsep":  [0.,                      util.noop],
    "opslewwt":  [0.,                      util.noop],
    "opslewti":  [360.,                    util.noop],
    "ophlimwt":  [0.,                      util.noop],
    "ophlimti":  [1800.,                   util.noop],
    "opha":      [0.,                      util.noop],
    "ophawid":   [0.,                      util.noop],
    "ophawt":    [1.,                      util.noop],
    "ophmaxdt":  [7200.,                   util.noop],
    "wrap24":    [1.,                      util.to_bool],
    "expt":      ["No description given.", util.noop],
    "expcode":   ["NUG",                   util.noop],
    "linepg":    [55.,                     util.noop],
    "tpref":     [-1.,                     util.noop],
    "ptdur":     [20.,                     util.noop],
    "precdate":  [1979.9,                  util.noop],
    "dovex":     [0.,                      util.to_bool],
    "vextest":   [1.,                      util.to_bool],
    "domka":     [1.,                      util.to_bool],
    "obstype":   ["NONE",                  util.upper],
    "doscans":   [[0., 0.],                util.noop],
    "optmode":   ["NONE",                  util.upper],
    "opdur":     [0.,                      util.multiply_by(parameter.onesec)],
    "opnosub":   [1.,                      util.to_bool],
    "opskip":    [0.,                      util.noop],
    "optslew":   [1.,                      util.noop],
    "optlowt":   [15.,                     util.noop],
    "ophasta":   ["PT",                    util.upper],
    "tapesync":  [1.,                      util.to_bool],
    "opprtlev":  [0.,                      util.noop],
    "opelprio":  [[0.] * 4,                util.noop],
    "maplim":    [[0.] * 4,                util.noop],
    "gridnr":    [20.,                     util.noop],
    "gridnt":    [36.,                     util.noop],
    "gridmin":   [25.,                     util.noop],
    "gridmax":   [250.,                    util.noop],
    "gridw0":    [0.,                      util.noop],
    "gridstep":  [3.,                      util.noop],
    "gridmeas":  ["COUNT",                 util.upper],
    "gridvla":   [1.,                      util.to_bool],
    "uvmfs":     [[1., 1.],                util.noop],
    "geoprt":    [-1.,                     util.noop],
    "geotries":  [20,                      util.noop],
    "geoback":   [100.,                    util.noop],
    "geoslew":   [1.,                      util.noop],
    "geoslow":   [40.,                     util.noop],
    "geosrep":   [100,                     util.noop],
    "geohiel":   [40,                      util.noop],
    "geolowel":  [23.,                     util.noop],
    "geosrcs":   [[],                      util.foreach(util.upper)],
    "version":   [0.,                      util.noop],
    "piname":    ["",                      util.noop],
    "address1":  ["",                      util.noop],
    "address2":  ["",                      util.noop],
    "address3":  ["",                      util.noop],
    "address4":  ["",                      util.noop],
    "phone":     ["",                      util.noop],
    "email":     ["",                      util.noop],
    "fax":       ["",                      util.noop],
    "obsphone":  ["",                      util.noop],
    "obsmode":   ["",                      util.noop],
    "note1":     ["",                      util.noop],
    "note2":     ["",                      util.noop],
    "note3":     ["",                      util.noop],
    "note4":     ["",                      util.noop],
    "correl":    ["",                      util.strip_upper],
    "coravg":    [[0., ""],                util.noop],
    "coravg2":   [[0., ""],                util.noop],
    "corchan":   [[0., 0.],                util.noop],
    "cornant":   [0.,                      util.noop],
    "corsrcs":   ["Not specified.",        util.upper],
    "corpol":    ["",                      util.upper],
    "corwtfn":   ["UNIFORM",               util.upper],
    "cortape":   ["",                      util.upper],
    "cordfmt":   ["FITS",                  util.upper],
    "corship1":  ["",                      util.noop],
    "corship2":  ["",                      util.noop],
    "corship3":  ["",                      util.noop],
    "corship4":  ["",                      util.noop],
    "cornote1":  ["",                      util.noop],
    "cornote2":  ["",                      util.noop],
    "cornote3":  ["",                      util.noop],
    "cornote4":  ["",                      util.noop],
    "rotpat":    [0.,                      util.noop],
    "focoff":    [[0.] * 20,               util.noop],
    "rotoff":    [[0.] * 20,               util.noop],
    "sumitem":   [[""] * 10,               util.foreach(util.upper)],
    # the parameter LST is both a boolean flag (implemented by floats) 
    # and a string for the LST station, the logic is implemented in times()
    "lst":       [1.0,                     util.noop],
    "tantsta1":  [[""] * maxsta,           util.foreach(util.upper)],
    "tantsta2":  [[""] * maxsta,           util.foreach(util.upper)],
    "prestart":  [parameter.unset,         util.multiply_by(parameter.onesec)],
    "minpause":  [parameter.unset,         util.multiply_by(parameter.onesec)],
    "debug":     [1.,                      util.to_bool],
    "overwrite": [1.,                      util.to_bool],
    "override":  [1.,                      util.to_bool],
    "nosetup":   [1.,                      util.to_bool],
    "plot":      [1.,                      util.to_bool],
    "pubplot":   [1.,                      util.to_bool],
    "scanexps":  [[],                      util.noop],
}

attribute_to_key = {
    "prescan": "prescan",
    "gap": "gap",
    "datapath": "datapath",
    "grabto": "grabto",
    "grabtime": "grabtime",
    "grabgap": "grabgap",
    "scnsrc": "source",
    "qual": "qual",
    "geolen": "geoseg",
    "centers": "centers",
    "tant1": "tant1",
    "tant2": "tant2",
    "annot": "comment",
    "caltime": "caltime",
    "ptslew": "ptslew",
    "pcal": "pcal",
    "focus": "focus",
    "rotation": "rotation",
    "opmiss": "opmiss",
    "scantag": "scantag",
    "crdline": "crdline",
    "dodown": "dodown",
    "preempt": "preempt",
    "dopeak": "peak",
    "point": "point",
    "higroup": "higroup",
    "freq": "freq",
    "bw": "bw",
    "crdnch": "crdnch",
    "crdch1": "crdch1",
    "crdsetch": "crdsetch",
    "crdfreq": "crdfreq",
    "crdbw": "crdbw",
    "dopsrc": "dopsrc",
    "dopcal": "dopcal",
    "lines": "linename",
    "vlamode": "vlamode",
    "vlapeak": "vlapeak",
    "vlaphs": "vlapsrc",
    "vlainteg": "vlainteg",
    "vlaptime": "vlaptime",
    "opminel": "opminel",
    "opmian": "opminant",
    "opminsep": "opminsep",
    "opslewwt": "opslewwt",
    "opslewti": "opslewti",
    "ophlimwt": "ophlimwt",
    "ophlimti": "ophlimti",
    "opha": "opha",
    "ophawid": "ophawid",
    "ophawt": "ophawt",
    "ophmaxdt": "ophmaxdt",
    "dur": "duration",
    "prestart": "prestart",
    "minpause": "minpause",
    "sazcol": "azcolim",
    "selcol": "elcolim",
    "scanexps": "scanexps",
}

def schin(stdin):
    global record_defaults, state_defaults

    catalog = ScanCatalog()

    year = np.zeros(shape=(catalog.maxscan,), dtype="int32")
    day = np.zeros(shape=(catalog.maxscan,), dtype="int32")
    stop = np.zeros(shape=(catalog.maxscan,), dtype="double")
    start = np.zeros(shape=(catalog.maxscan,), dtype="double")
    mjd1 = 0
    if state_defaults["schedule"][0] == "":
        input_ = stdin
    else:
        try:
            input_ = open(util.expand_file_name(state_defaults["schedule"][0]),
                          "r")
        except Exception as e:
            s.wlog(1, str(e))
            s.errlog("SCHFILES: Problem opening schedule file {}".format(
                state_defaults["schedule"][0]))
    # the iterator might change, from file/stdin to (another) file,
    input_iterator = key.KeyfileIterator(input_, 
                                         record_defaults, state_defaults)
    
    s.schn1.nsta = 0
    s.schsat.nsat = 0
    marker = -1
    s.schsta.msta = 0
    s.schsou.msrc = 0
    s.schsf.nsetf = 0
    s.schn1.scan1 = 1
    s.schsou.ngeo = 0
    gotsat = False
    doinit = True
    s.schn1.dwells = False
    dostwarn = True
    s.schcon.dovex = True
    gotvex = False
    s.schcon.coverlet = False
    s.schn5.allvlba = True
    s.schsou.anygeo = False
    s.schn1.fuzzy = False
    s.schn1.gotvlba = False
    s.schn1.gotpreem = False
    
    s.schn2a.srcnum.fill(0)
    s.schn2a.idopsrc.fill(0)
    s.schn3.ivlaphs.fill(0)
    s.schn3.vlainteg.fill(0)
    
    s.schsou.sused.fill(False)
    s.schsou.usedrec.fill(False)
    s.schsou.usedphs.fill(False)
    s.schsou.usedcent.fill(False)
    s.schn1.srcatn.fill(0)
    s.schsou.srlstn.fill(0)
    s.schsou.didndop.fill(0)
    s.schsou.dopped.fill(False)
    s.schcsc.csused.fill(b' ')
    s.schsou.planet.fill(False)
    s.schsou.satel.fill(False)
    s.schsou.satn.fill(0)

    s.schn2a.nintent = 0
    s.schc2c.intent.fill(b' ' * s.schc2c.intent.itemsize)
    s.schn2a.nscint.fill(0)
    s.schn2a.iscint.fill(0)

    restart = False

    index = 0
    while True:
        try:
            record = next(input_iterator)
            # a weird one, only dwell[0]'s default does not carry over
            state_defaults["dwell"][0][0] = parameter.unset
            # make a new scanexps overwrite the old list, 
            # instead of the default behaviour of slicing into it
            old_scanexps = state_defaults["scanexps"][0]
            state_defaults["scanexps"][0] = []
            if restart:
                old_values, old_present = values, present
                values, present = util.merge_record_with_defaults(
                    record, record_defaults, state_defaults)
                values.update({k: old_values[k] for k in keep_for_next
                               if k not in present})
                present.update(old_present & keep_for_next)
                restart = False
            else:
                values, present = util.merge_record_with_defaults(
                    record, record_defaults, state_defaults)
            if "scanexps" not in present:
                state_defaults["scanexps"][0] = old_scanexps
                values["scanexps"] = old_scanexps
            
            if "exit" in present:
                if index == 0:
                    s.wlog(1, "SCHIN:  EXIT requested.  Shutting down.")
                    s.delscr(False)
                    sys.exit(0)
                break

            s.schcon.debug = values["debug"]
            s.schcon.overwrit = values["overwrite"]
            s.schcon.override = values["override"]

            s.schsco.msgfile = util.resize_string(util.expand_file_name(
                values["msgfile"]), s.schsco.msgfile.itemsize, "msgfile")

            if s.schcon.debug and (index < 3):
                s.wlog(0, "DIVERT: Starting")

            s.schcon.freqlist = values["freqlist"]
            if "freqlist" in present:
                s.schsco.freqfile = util.resize_string(util.expand_file_name(
                    values["freqfile"]), s.schsco.freqfile.itemsize, "freqfile")
                getfreq()
                s.wlog(1, "DIVERT:   Frequency table written.  Stopping.")
                sys.exit(0)

            s.schcon.noset = values["nosetup"]
            s.schcon.plot = values["plot"]
            s.schcon.pubplot = values["pubplot"]

            schedule = util.expand_file_name(values["schedule"])
            if (schedule != "") and \
               ((input_iterator.input_ is stdin) or 
                (input_iterator.input_.name != schedule)):
                if s.schcon.debug:
                    s.wlog(0, "SCHFILES:  About to open {}".format(schedule))
                if input_iterator.input_ is not stdin:
                    input_iterator.input_.close()
                try:
                    input_ = open(schedule, "r")
                    input_iterator = key.KeyfileIterator(
                        input_, record_defaults, state_defaults)
                    restart = True
                    continue # read next record from new iterator
                except Exception as e:
                    s.wlog(1, str(e))
                    s.errlog("SCHFILES: Problem opening schedule file {}".\
                             format(schedule))
            gotsat, restart = schfiles(input_iterator, stdin, 
                                       values, present, gotsat)
            input_iterator.set_defaults(record_defaults, state_defaults)
            if restart:
                continue

            if index > catalog.maxscan:
                s.errlog("SCHIN: Too many scans, maximum {}".format(
                    catalog.maxscan))
            entry = catalog.entries[index]

            s.schsco.dosta = util.resize_string(
                values["dosta"], s.schsco.dosta.itemsize, "dosta")
            if (values["dosta"] != "ALL") and dostwarn:
                s.wlog(0, "SCHIN:  DOSTA specified as {}".format(
                    values["dosta"]))
                s.wlog(0, "        Some stations may be skipped.")
                dostwarn = False

            # nopeak overwrites peak and its default
            if "nopeak" in present:
                state_defaults["peak"][0] = values["peak"] = -1.

            if ("scanexps" in present) and \
               ("NONE" in (v.upper() for v in values["scanexps"])):
                state_defaults["scanexps"][0] = values["scanexps"] = []
            
            entry.set_keyin_values(values, attribute_to_key)

            mjd1 = gettim(values, catalog.entries, index,
                          start, stop, day, year, mjd1)
            gotvex = getsta(stdin, values, index, gotvex, mjd1)

            if entry.scnsrc == "":
                s.errlog("SCHIN: Need source name - blank specified.")

            if entry.geolen > 0:
                s.schsou.anygeo = True
                entry.geoiscn = index + 1

            pcal_map = {
                "OFF": "off",
                "1MHZ": "1MHz",
                "5MHZ": "5MHz"
            }
            if entry.pcal in pcal_map.keys():
                entry.pcal = pcal_map[entry.pcal]
            elif entry.pcal != "":
                s.errlog("SCHIN: Invalid PCAL ({}) specified in scan {}".\
                         format(entry.pcal, index+1))
            if entry.preempt == "EXTRA":
                s.schn1.fuzzy = True
            elif entry.preempt == "--":
                if entry.geolen > 0.:
                    entry.preempt = "NO"
                else:
                    entry.preempt = "OK"
            if entry.preempt != "OK":
                s.schn1.gotpreem = True

            gintent(values["intents"], catalog.entries, index)
            
            # toggle pairs
            if index == 0:
                get_default = lambda attribute: False
            else:
                prev_entry = catalog.entries[index-1]
                get_default = lambda attribute: getattr(prev_entry, attribute)
            for attribute, key1, key2 in (
                    ("notsys", "notsys", "tsys"),
                    ("norec", "norecord", "record"),
                    ("pntvlba", "ptvlba", "noptvlba"),
                    ("tanvlba", "tavlba", "notavlba"),
                    ("dopn3db", "pn3db", "nopn3db")):
                setattr(entry, attribute,
                        toggle(values, present, key1, key2, 
                               get_default(attribute)))
            
            s.schcon.autopeak = values["autopeak"]
            s.schcon.pkwatch = values["pkwatch"]
            s.schsco.peakfile = util.resize_string(util.expand_file_name(
                values["peakfile"]), s.schsco.peakfile.itemsize, "peakfile")

            entry.setnum = SetupFileCatalog.extend_with(
                util.expand_file_name(values["setup"]), 
                " SCHIN: Too many setup files. ")
            
            infdb(values, present, catalog.entries, index)
            invla(values, present, catalog.entries, index, 1)
            
            index, marker = schrep(values, catalog.entries, index, marker,
                                   start, stop, day, year)

            if "nchan" in present:
                s.wlog(1, "SCHIN: NCHAN in main schedule now ignored.")

            if s.schcon.debug:
                s.wlog(0, "SCHIN: Finished reading scan: {}".format(index+1))

            index += 1
        except StopIteration:
            break
    # end of scan reading loop
    
    if input_iterator.input_ is not stdin:
        input_iterator.input_.close()

    s.schn1.nscans = index
    s.schn1.scanl = s.schn1.nscans
    if index <= 0:
        s.errlog("SCHIN: No input scans")

    s.schn1.wrap24 = values["wrap24"]
    if s.schn1.wrap24:
        for from_ in range(s.schn1.scanl):
            if from_ > 0:
                if (start[from_] != parameter.unset) or \
                   (stop[from_] != parameter.unset):
                    s.errlog("SCHIN: Do not use START or STOP times "
                             "after scan 1 with WRAP24. See scan: {}".\
                             format(from_+1))
            to = from_ + s.schn1.nscans
            scndup(to, from_, False, "SCHIN", use_direct_access=False)
            catalog.entries[to].annot = catalog.entries[from_].annot
            start[to] = parameter.unset
            stop[to] = parameter.unset
            day[to] = day[from_]
            year[to] = year[from_]
        s.schn1.scanl = 2 * s.schn1.nscans
        s.schn1.nscans = s.schn1.scanl
        index *= 2
    
    s.schc1.expt = util.resize_string(values["expt"], s.schc1.expt.itemsize, 
                                      "expt")
    s.schc1.expcode = util.resize_string(values["expcode"], 
                                         s.schc1.expcode.itemsize, "expcode")
    s.schcon.linepg = values["linepg"]
    s.schn5.tpref = values["tpref"]
    s.schn4.ptdur = values["ptdur"]
    s.schcon.precdate = values["precdate"]

    s.schcon.dovex = (values["dovex"] or gotvex)
    s.schcon.vextest = values["vextest"]

    s.schcon.domka = values["domka"]

    s.schsco.ephfile = util.resize_string(util.expand_file_name(
        values["ephfile"]), s.schsco.ephfile.itemsize, "ephfile")

    s.chkcode(s.schc1.expcode)

    obstype = values["obstype"]
    if obstype[:4] == "MARK":
        obstype = "MK" + obstype[4:]
    s.schsco.obstyp = util.resize_string(obstype, s.schsco.obstyp.itemsize, 
                                         "obstype")
    s.schn1.mark2 = (obstype == "MKII")
    s.schn1.vlbitp = (obstype in ("VLBA", "MKIII", "VLBI", "MKIV"))
    s.schn1.vlaonly = (obstype == "VLA")
    s.schn1.notape = (obstype in ("VLA", "NONE", "PTVLBA", "CONFIG"))
    s.schn1.config = (obstype == "CONFIG")
    if not (s.schn1.mark2 or s.schn1.vlbitp or s.schn1.vlaonly or 
            obstype in ("NONE", "PTVLBA", "CONFIG")):
        s.wlog(1, " SCHIN: Invalid OBSTYPE: {}".format(obstype))
        s.errlog(" SCHIN: OBSTYPE must be MKII, MKIII, VLBA, MKIV, "
                 "VLBI, VLA, NONE, or CONFIG")

    s.schn1.doscans = values["doscans"]
    if ((s.schn1.doscans[0] == 0) != (s.schn1.doscans[1] == 0)):
        s.errlog("If using DOSCANS, specify both!")

    s.schsco.optmode = util.resize_string(values["optmode"], 
                                          s.schsco.optmode.itemsize, "optmode")
    s.schcon.opdur   = values["opdur"]
    s.schcon.opnosub = values["opnosub"]
    s.schcon.opskip  = values["opskip"]
    s.schcon.optslew = values["optslew"]
    s.schcon.optlowt = values["optlowt"]
    s.schsco.ophasta = util.resize_string(values["ophasta"], 
                                          s.schsco.ophasta.itemsize, "ophasta")
    s.schcon.tapesync = values["tapesync"]
    s.schcon.opprtlev = values["opprtlev"]

    s.schcon.opelprio = values["opelprio"]
    s.schcon.maplim = values["maplim"]

    s.schcon.gridnr   = values["gridnr"]
    s.schcon.gridnt   = values["gridnt"]
    s.schcon.gridmin  = values["gridmin"]
    s.schcon.gridmax  = values["gridmax"]
    s.schcon.gridw0   = values["gridw0"]
    s.schcon.gridstep = values["gridstep"]
    s.schsco.gridmeas = util.resize_string(values["gridmeas"], 
                                           s.schsco.gridmeas.itemsize, 
                                           "gridmeas")
    s.schcon.gridvla  = values["gridvla"]
    s.schcon.gridused = False
    s.schcon.nmfs = values["uvmfs"][0]
    s.schcon.mfsrat = values["uvmfs"][1]

    mgeo = s.schsou.geosrci.shape[0]
    geosrcs = [util.resize_string(src, s.schcsc.geosrc.itemsize, "geosrcs") 
               for src in values["geosrcs"]]
    s.schsou.ngeo = len(geosrcs[:mgeo])
    for i, src in enumerate(geosrcs):
        s.schcsc.geosrc[i] = src

    if s.schsou.anygeo and s.schsou.ngeo == 0:
        s.errlog("Geodetic segments requested, but no GEOSRCS given.")

    if s.schsou.anygeo:
        for entry in catalog.entries:
            if entry.scnsrc == "GEOSEG":
                entry.scnsrc = geosrcs[0]
    
    s.schsou.geoprt   = values["geoprt"]
    s.schsou.geotries = values["geotries"]
    s.schsou.geoback  = values["geoback"]
    s.schsou.geoslew  = values["geoslew"]
    s.schsou.geoslow  = values["geoslow"]
    s.schsou.geosrep  = values["geosrep"]
    s.schsou.geohiel  = values["geohiel"]
    s.schsou.geolowel = values["geolowel"]

    getcov(values)
    getcor(values)

    s.schn4.rotpat = values["rotpat"]
    s.schn4.foc = values["focoff"]
    s.schn4.rot = values["rotoff"]

    if "tapefile" in present:
        s.errlog("TAPEFILE given but tape initialization no longer "
                 "supported by SCHED")
    s.tptpns()

    for i, sumitem in enumerate(values["sumitem"]):
        s.schsco.sumitem[i] = util.resize_string(sumitem, 
                                                 s.schsco.sumitem.itemsize, 
                                                 "sumitem")
    if values["sumitem"][0] == "":
        s.schsco.sumitem[0] = "ELA".ljust(s.schsco.sumitem.itemsize)
        s.schsco.sumitem[1] = "DWELL".ljust(s.schsco.sumitem.itemsize)

    invla(values, set(), catalog.entries, index, 2) # index is dummy in case 2

    if s.schcon.debug:
        s.wlog(0, "SCHIN: About to read catalogs.")

    catalog.write(range(index))
    
    times(values["lst"], start, stop, day, year)
    
    s.schcsc.srcfile = util.resize_string(util.expand_file_name(
        values["srcfile"]), s.schcsc.srcfile.itemsize, "srcfile")
    s.schcsc.srcfile2 = util.resize_string(util.expand_file_name(
        values["srcfile2"]), s.schcsc.srcfile2.itemsize, "srcfile2")

    sttant(values["tantsta1"], values["tantsta2"])

    s.schsco.freqfile = util.resize_string(util.expand_file_name(
        values["freqfile"]), s.schsco.freqfile.itemsize, "freqfile")
    s.schcon.freqlist = values["freqlist"]

    if "autotape" in present:
        s.wlog(1, "SCHIN:  Obsolete parameter AUTOTAPE given.  Ignored.")
    if "tape" in present:
        s.wlog(1, "'SCHIN:  Obsolete parameter TAPE given.  Ignored.")
    if "fastfor" in present:
        s.wlog(1, "SCHIN:  Obsolete parameter FASTFOR given.  Ignored.")
    if "reverse" in present:
        s.wlog(1, "SCHIN:  Obsolete parameter REVERSE given.  Ignored.")



from . import scndup
from ..catalog import ScanCatalog, StationCatalog, SourceCatalog
from ..util import f2str

import schedlib as s

import numpy as np

skip = np.empty(dtype=int, shape=(SourceCatalog.maxsource,))
missed = 0
def optskd(last_scan_index, k_scan, scan_index):
    global skip, missed

    if k_scan > s.schn1.nscans:
        return False, False, True
    
    if k_scan != scan_index:
        scndup(scan_index, k_scan, True, "OPTSKD")

    scans = ScanCatalog().direct_access_entries
    stations = StationCatalog().used(use_direct_access=True)
    scan = scans[scan_index - 1]
    
    adjust = (scan.duronly == 1)

    if scan_index == 1:
        skip.fill(s.schcon.opskip)
        missed = 0

    if not adjust:
        approx_time = scan.startj
    else:
        approx_time = s.schn1.tfirst
        for i, station in enumerate(stations):
            if station.stascn[scan_index - 1] and (last_scan_index[i] != 0):
                approx_time = max(approx_time, 
                                  scans[last_scan_index[i] - 1].stopj + scan.gap)

    n_good = 0
    n_prio = 0
    begin_time = s.schn1.tfirst
    ok_sta = np.full(fill_value=False, dtype=bool, shape=(len(stations,)))
    for i, station in enumerate(stations):
        if station.stascn[scan_index - 1]:
            last_time, avail_time = s.stageo(scan_index, i + 1, approx_time, 
                                             last_scan_index[i], "OPTSKD")
            if (f2str(station.up1[scan_index - 1]) == "") and \
               (f2str(station.up2[scan_index - 1]) == "") and \
               (station.el1[scan_index - 1] > scan.opminel) and \
               (station.el2[scan_index - 1] > scan.opminel):
                ok_sta[i] = True
                n_good += 1
                if last_scan_index[i] != 0:
                    avail_time = max(
                        avail_time, 
                        scans[last_scan_index[i] - 1].stopj + scan.gap)
                    begin_time = max(begin_time, avail_time)
                opelprio = s.schcon.opelprio
                if (opelprio[0] < station.el1[scan_index - 1] < opelprio[1]) or \
                   (opelprio[2] < station.el1[scan_index - 1] < opelprio[3]):
                    n_prio += 1
    
    if (not adjust) or (begin_time == s.schn1.tfirst):
        begin_time = approx_time

    source_index = scan.srcnum

    keep = (n_good >= max(scan.opmian, 1)) and \
           ((skip[source_index - 1] >= s.schcon.opskip) or (n_prio >= 1))
    if f2str(s.schsco.obstyp) not in {"NONE", "PTVLBA"}:
        keep = keep or ((n_good >= 1) and 
                        (scan.pntvlba or scan.tanvlba or scan.dopn3db))
    keep = keep and (missed >= scan.opmiss)
    
    if keep:
        skip[source_index - 1] = 0
        if adjust:
            scan.startj = begin_time
            scan.stopj = begin_time + scan.dur

        for i, station in enumerate(stations):
            station.stascn[scan_index - 1] = (ok_sta[i] or s.schcon.opnosub)

        missed = 0
    else:
        if n_good >= scan.opmian:
            skip[source_index - 1] += 1

        for station in stations:
            station.stascn[scan_index - 1] = False

        missed += 1
        
    return adjust, keep, False



#!/usr/bin/env python3

import setuptools
import numpy.f2py as f2py
from numpy.distutils.core import setup, Extension
from distutils.command.sdist import sdist

import platform
import argparse
import sys
import os

parser = argparse.ArgumentParser(
    description="Setup script for pySCHED. Builds a list of source files "
    "depending on the system architecture. Other arguments will be forwarded "
    "to setup.",
    add_help=False)
parser.add_argument("--schedarch",  
                    choices=["linux", "unix", "sun", "hp"],
                    default=None, 
                    help="operating system for machine dependent fortran "
                    "routines. Defaults to linux.")

args, other_args = parser.parse_known_args()
sys.argv[1:] = other_args

if args.schedarch is None:
    arch = "linux"
    system = platform.system()
    if system not in ["Linux", "MacOS"]:
        print("Warning: unknown architecture '{}'. Will fall back to '{}'.".\
              format(system, arch))
else:
    arch = args.schedarch

# source files to compile
sources = """
Cit/geoid.f
Cit/geoxyz.f
Cit/julda.f
Cit/len1.f
Cit/pangle.f
Cit/tdatecw.f
Cit/tform.f
Cit/upcase.f
Jpl/const.f
Jpl/findt.f
Jpl/fsizer2.f
Jpl/interp.f
Jpl/jplep2.f
Jpl/jplpo2.f
Jpl/jplver.f
Jpl/pleph.f
Jpl/pvobs.f
Jpl/split.f
Jpl/state.f
Jpl/transp.f
Plot/plbeam.f
Plot/plfft.f
Plot/pltran.f
Plot/plver.f
Plot/plweig.f
Plotstub/pgqinf.f
Satstub/satep.f
Satstub/sattle.f
Sched/accsrc.f
Sched/addgeo.f
Sched/addpeak.f
Sched/antpos.f
Sched/autodown.f
Sched/azelxyew.f
Sched/azelxyns.f
Sched/badlo.f
Sched/bbcalt.f
Sched/bbccdas.f
Sched/bbcdbbc.f
Sched/bbcgeo.f
Sched/bbckvsr.f
Sched/bbclba.f
Sched/bbcm4.f
Sched/bbcrdbe.f
Sched/bbcvlba.f
Sched/bbcvs2.f
Sched/bbcwidar.f
Sched/chars.f
Sched/chkcdas.f
Sched/chkcode.f
Sched/chkcor.f
Sched/chkdbbc.f
Sched/chkdbfq.f
Sched/chkdisk.f
Sched/chkfirm.f
Sched/chkgdar.f
Sched/chkif.f
Sched/chkrdbe.f
Sched/chkrdfq.f
Sched/chksc1.f
Sched/chkscn.f
Sched/chkset.f
Sched/chksfil.f
Sched/chkspd.f
Sched/chkvdar.f
Sched/chkvdifx.f
Sched/chkvla.f
Sched/chkvlba.f
Sched/chkwidar.f
Sched/cordef.f
Sched/corlst.f
Sched/corsoc.f
Sched/crdwrt.f
Sched/defaults.f
Sched/defset.f
Sched/delscr.f
Sched/dequal.f
Sched/diskpos.f
Sched/dopcrd.f
Sched/dopfq.f
Sched/dwcase.f
Sched/errlog.f
Sched/errset.f
Sched/fcompare.f
Sched/fileopen.f
Sched/flags.f
Sched/flagwrt.f
Sched/flush.f
Sched/fmatch.f
Sched/fmtmkiii.f
Sched/fmtmkiv.f
Sched/fmtpick.f
Sched/fmts2.f
Sched/fmtvlba.f
Sched/frchar.f
Sched/fsfreq.f
Sched/fsmatch.f
Sched/fspread.f
Sched/fsvlba.f
Sched/geochk.f
Sched/geomake.f
Sched/geoqual.f
Sched/getcrdn.f
Sched/getfset.f
Sched/getpairs.f
Sched/getsun.f
Sched/glstday.f
Sched/gmkscn.f
Sched/gnset.f
Sched/haavai.f
Sched/halim.f
Sched/harmwarn.f
Sched/headpos.f
Sched/horchk.f
Sched/ifdbbc.f
Sched/inarray.f
Sched/jplgot.f
Sched/listfreq.f
Sched/lsqfit.f
Sched/lst2ut.f
Sched/lstfreq.f
Sched/makeptg.f
Sched/makescn.f
Sched/makeseg.f
Sched/maxbas.f
Sched/mtaltaz.f
Sched/mtequat.f
Sched/mtxyew.f
Sched/mtxyns.f
Sched/omscor.f
Sched/omsfreq.f
Sched/omsout.f
Sched/omsset.f
Sched/omssrc.f
Sched/omssta.f
Sched/optcells.f
Sched/optcsar.f
Sched/optcspt.f
Sched/optcsub.f
Sched/opthas.f
Sched/opthiel.f
Sched/optnone.f
Sched/optsch.f
Sched/optskd.f
Sched/opttim.f
Sched/optupt.f
Sched/pcalfq.f
Sched/phint.f
Sched/pkfinish.f
Sched/pn3db.f
Sched/protect.f
Sched/prtfreq.f
Sched/prtsch.f
Sched/prtscn.f
Sched/prtset.f
Sched/ptpat.f
Sched/ptvlba.f
Sched/ran5.f
Sched/rdbelevt.f
Sched/rdbemtch.f
Sched/recctl.f
Sched/reconfig.f
Sched/resync.f
Sched/rotvlba.f
Sched/sameset.f
Sched/satgot.f
Sched/sattim.f
Sched/sbhours.f
Sched/sbpair.f
Sched/scanid.f
Sched/sch24.f
Sched/schday.f
Sched/schgeo.f
Sched/schopt.f
Sched/schpre.f
Sched/schsum.f
Sched/schtim.f
Sched/scndup.f
Sched/scngeo.f
Sched/scnrange.f
Sched/scramble.f
Sched/sdopincr.f
Sched/setband.f
Sched/setbbc.f
Sched/setchan.f
Sched/setcop.f
Sched/setdefs.f
Sched/setexpnd.f
Sched/setfcat.f
Sched/setfirm.f
Sched/setfmt.f
Sched/setform.f
Sched/setfreq.f
Sched/sethw1.f
Sched/setnorec.f
Sched/setrec.f
Sched/setstdef.f
Sched/settps.f
Sched/settrk.f
Sched/setusyn.f
Sched/sfinfo.f
Sched/shortn.f
Sched/sidtim.f
Sched/slew.f
Sched/socdef.f
Sched/srcflg.f
Sched/srcloc.f
Sched/srclst.f
Sched/srcwrt.f
Sched/srfinish.f
Sched/srinsert.f
Sched/srlist.f
Sched/srlpre.f
Sched/stafiles.f
Sched/stafrd.f
Sched/stageo.f
Sched/stalst.f
Sched/stano.f
Sched/stauv.f
Sched/stmsg.f
Sched/stsum.f
Sched/sumdat.f
Sched/sumdesc.f
Sched/sumope.f
Sched/sumscn.f
Sched/sunpos.f
Sched/sunwarn.f
Sched/suvopt.f
Sched/tavlba.f
Sched/timej.f
Sched/tpmfix.f
Sched/tpsum.f
Sched/tptpns.f
Sched/twohdset.f
Sched/uvopt.f
Sched/uvqual.f
Sched/v2dout.f
Sched/versched.f
Sched/vexout.f
Sched/vlascns.f
Sched/vlba.f
Sched/vlbabws.f
Sched/vlbachar.f
Sched/vlbadk.f
Sched/vlbaend.f
Sched/vlbaini.f
Sched/vlbaint.f
Sched/vlbareal.f
Sched/vlbast.f
Sched/vlbastop.f
Sched/vlbasu.f
Sched/vsopwrt.f
Sched/wlog.f
Sched/wrap.f
Sched/wrapzone.f
Sched/wrtcov.f
Sched/wrtfreq.f
Sched/wrtmsg.f
Sla/amp.f
Sla/ampqk.f
Sla/calyd.f
Sla/cldj.f
Sla/clyd.f
Sla/cs2c.f
Sla/dat.f
Sla/dcc2s.f
Sla/dcs2c.f
Sla/deuler.f
Sla/dimxv.f
Sla/djcl.f
Sla/dmxm.f
Sla/dmxv.f
Sla/dranrm.f
Sla/dsep.f
Sla/dsepv.f
Sla/dt.f
Sla/dvdv.f
Sla/dvn.f
Sla/dvxv.f
Sla/earth.f
Sla/ecor.f
Sla/epb2d.f
Sla/epj.f
Sla/eqeqx.f
Sla/evp.f
Sla/fk45z.f
Sla/fk524.f
Sla/fk54z.f
Sla/gmst.f
Sla/map.f
Sla/mappa.f
Sla/mapqk.f
Sla/mapqkz.f
Sla/nut.f
Sla/nutc.f
Sla/pm.f
Sla/prebn.f
Sla/prec.f
Sla/preces.f
Sla/precl.f
Sla/prenut.f
Sla/rvlsrk.f
Sla/vdv.f
Vex/chk4dar.f
Vex/chkjive.f
Vex/chkv4dar.f
Vex/vxcfan.f
Vex/vxcfbb.f
Vex/vxcfda.f
Vex/vxcffq.f
Vex/vxcfhp.f
Vex/vxcfif.f
Vex/vxcfph.f
Vex/vxcfpo.f
Vex/vxcfrl.f
Vex/vxcfsi.f
Vex/vxcftr.f
Vex/vxcovr.f
Vex/vxdefs.f
Vex/vxfqvx.f
Vex/vxgtst.f
Vex/vxmode.f
Vex/vxnhds.f
Vex/vxnman.f
Vex/vxnmbb.f
Vex/vxnmda.f
Vex/vxnmfq.f
Vex/vxnmhp.f
Vex/vxnmif.f
Vex/vxnmph.f
Vex/vxnmpo.f
Vex/vxnmpr.f
Vex/vxnmrl.f
Vex/vxnmsi.f
Vex/vxnmtr.f
Vex/vxnmxx.f
Vex/vxrl16.f
Vex/vxrl8.f
Vex/vxs2df.f
Vex/vxs2md.f
Vex/vxs2vl.f
Vex/vxsch.f
Vex/vxschk.f
Vex/vxscns.f
Vex/vxsort.f
Vex/vxstky.f
Vex/vxstli.f
Vex/vxstnm.f
Vex/vxsudt.f
Vex/vxtels.f
Vex/vxton2.f
Vex/vxtone.f
Vex/vxtrafq.f
Vex/vxtraif.f
Vex/vxtramd.f
Vex/vxtran.f
Vex/vxtraph.f
Vex/vxunql.f
Vex/vxvbrx.f
Vex/vxvers.f
Vex/vxwran.f
Vex/vxwrbb.f
Vex/vxwrda.f
Vex/vxwrex.f
Vex/vxwrfq.f
Vex/vxwrgl.f
Vex/vxwrhp.f
Vex/vxwrif.f
Vex/vxwrmd.f
Vex/vxwrph.f
Vex/vxwrpo.f
Vex/vxwrpr.f
Vex/vxwrrl.f
Vex/vxwrsi.f
Vex/vxwrst.f
Vex/vxwrsu.f
Vex/vxwrt.f
Vex/vxwrtr.f 
pysched/openwrap.f 
pysched/tformwrp.f 
pysched/ran5wrap.f 
pysched/verwrap.f 
""".split()

if arch in ["hp", "linux", "sun", "unix"]:
    sources += """
Cit/sys_unix/envir.f
Cit/sys_unix/krdlin.f
Cit/sys_unix/prognm.f
Cit/sys_unix/putout.f
Cit/sys_unix/schdefs.f
Cit/sys_unix/symsub.f
Cit/sys_unix/tsttty.f
Cit/sys_unix/vmshlp.f
""".split()

if arch == "linux":
    sources += """
Cit/sys_linux/error.f
Cit/sys_linux/gerror.c
Cit/sys_linux/idate.c
Cit/sys_linux/isatty.c
Cit/sys_linux/vlbope.f
""".split()
elif arch in ["sun", "hp"]:
    sources += """
Cit/sys_sun/error.f
Cit/sys_sun/vlbope.f
""".split()

# fortran functions to create a python interface for
functions = """
accsrc 
autodown 
bbccdas 
bbcgeo 
bbckvsr 
bbclba 
bbcm4 
bbcrdbe 
bbcvlba 
bbcwidar 
chk4dar 
chkcdas 
chkcode 
chkcor 
chkdbfq 
chkdisk 
chkfirm 
chkjive 
chkrdbe 
chkv4dar
chkvdar 
chkvgdar 
chkvla 
chkvlba  
chkwidar 
chksc1 
chkscn 
chksfil 
chkspd 
chkvdifx 
cordef 
delscr 
diskpos 
dopfq 
errlog 
error 
errset 
flags 
fluxh 
geoqual 
geoxyz 
getpairs 
getsun 
gnset 
haavai 
halim 
jplver 
listfreq 
lst2ut 
makeptg 
omsout 
openwrap 
optsch 
opttim 
plbeam 
plver 
pkfinish 
prtscn 
putout 
ran5wrap 
recctl 
sattim 
sbpair 
sch24 
schgeo 
schpre 
schsum 
schtim 
scngeo 
scnrange 
setchan 
setfcat 
setfirm 
setfreq 
sethw1 
setrec 
setstdef 
settps 
settrk 
setusyn 
sdopincr 
sfinfo 
sidtim 
sla_calyd 
sla_cldj 
sla_dat 
socdef 
srcflg 
srfinish 
srinsert 
srlist 
stafiles 
stageo 
stano 
stauv 
stmsg 
sunpos 
tdatecw 
tformwrp 
timej 
tptpns 
v2dout 
verwrap 
vexout 
vlascns 
vsopwrt 
vxs2md 
vxton2 
wlog 
wrtmsg
""".split()

includes = """
Cit/rdcat.inc
Plot/beam.inc
Plot/plot.inc
Plot/proj.inc
Plot/sched.inc
Plot/schset.inc
Plot/srlist.inc
Sat/sched.inc
Sched/plot.inc
Sched/rdcat.inc
Sched/sched.inc
Sched/schfreq.inc
Sched/schpeak.inc
Sched/schset.inc
Sched/srlist.inc
Sched/vxlink.inc
Vex/sched.inc
Vex/schset.inc
Vex/vxlink.inc
""".split()

# include catalogs and setups directory contents
data_files = [(dir_, [os.path.join(dir_, file_) for file_ in files])
              for data_dir in ("catalogs", "setups")
              for dir_, _, files in os.walk(data_dir)]

# F2py assumes a Fortran dialect from the file name. For F77, it only accepts
# comments starting with a 'C'. But Sched uses '!' for comments too. This hack
# forces f2py to scan for those too.
f2py.crackfortran.is_f_file = lambda _: False

this_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_dir, "README.md")) as f:
    long_description = f.read()

extension = Extension(
    name="schedlib",
    sources=["src/" + s for s in sources + includes],
    f2py_options=["only:"] + functions + [":"])
setup(
    cmdclass={"sdist": sdist},
    name="pythonSCHED",
    version="1.15.0",
    author="Bob Eldering",
    author_email="eldering@jive.eu",
    description="Python extension of NRAO's VLBI scheduling program SCHED "
    "(see http://www.aoc.nrao.edu/~cwalker/sched/)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jive-vlbi/sched",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    data_files=data_files,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Fortran",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy"],
    scripts=["src/sched.py"],
    ext_modules=[extension],
    python_requires=">=3.6",
    setup_requires=["numpy>=1.16",
                    "setuptools>=24.3"],
    install_requires=["numpy>=1.16",
                      # on some systems pyqt 5.14 tries to build from source 
                      # and fails
                      "pyqt5<5.14", 
                      "matplotlib>=3",
                      "formlayout",
                      "astropy",
                      "bottle",
                      "gitpython",
                      "requests"
                  ]
)



from . import parameter, scndup

import schedlib as s

group = None
repeat = None

def schrep(values, entries, index, marker, start, stop, day, year):
    global group, repeat
    if s.schcon.debug:
        s.wlog(0, "SCHREP starting.")
    # if we get a repeat request, store it such that we can execute it when
    # the requested number of scan (group) has been parsed
    if marker == -1:
        repeat = int(values["repeat"])
        group = int(values["group"])
        if repeat != 0:
            marker = index
    elif values["repeat"] > 1:
        s.errlog("SCHREP: Cannot do nested or overlapping loops.")

    if s.schcon.debug:
        s.wlog(0, "SCHREP: Scan - {}  Repeat and group: {} {}".format(
            index+1, repeat, group))

    if (marker != -1) and (index + 1 >= group + marker):
        for r in range(repeat-1):
            for g in range(group):
                index += 1
                scndup(index, marker, False, "SCHREP", use_direct_access=False)
                start[index] = parameter.unset
                stop[index] = parameter.unset
                day[index] = day[marker]
                year[index] = year[marker]

                marker += 1

        marker = -1
    
    return index, marker



from . import chkdbbc, check_emerlin
from ..catalog import SetupCatalog, StationCatalog, FrequencyCatalog
from ..util import f2str

import schedlib as s

import numpy as np

import math

def freq_range(channel):
    # rounds to multiples of 10 (copied from SCHED)
    bandwidth = math.floor(channel.bbfilt / 10) * 10
    if channel.netside == "L":
        return (channel.freqref - bandwidth, channel.freqref)
    else:
        return (channel.freqref, channel.freqref + bandwidth)
    

itout = 1
warn2cm = True
def chkset(ks):
    global itout, warn2cm
    setup_entry = SetupCatalog().read()[ks-1]
    station_catalog = StationCatalog()
    station_catalog.read()
    station_catalog.read_non_scan_scheduled_attributes()
    station_entry = station_catalog.entries[setup_entry.isetsta-1]
    if s.schcon.debug:
        s.wlog(0, "CHKSET: Starting on {} in:".format(setup_entry.setsta[0]))
        s.wlog(0, "        {}".format(setup_entry.setname))

    errs = False
    sampwarn = True

    if s.schn1.notape and (setup_entry.format != "NONE"):
        obstype = f2str(s.schsco.obstyp)
        s.wlog(itout, "CHKSET: *** WARNING - OBSTYPE={} but setup has "
               "FORMAT={}".format(obstype, setup_entry.format))
        s.wlog(itout, "        No tapes will be recorded with this OBSTYPE.")
        if obstype == "NONE":
            s.wlog(itout, "        NONE is the default OBSTYPE. "
                   "Did you forget to specify it?")
        else:
            s.wlog(itout, "        Was this intended?")
        itout = 0
    
    if not s.schn1.vlaonly:
        if s.schn1.mark2 and (setup_entry.format != "MARKII"):
            s.wlog(1, "CHKSET: For OBSTYP=MKII, FORMAT must be MARKII, not: "
                   "{}".format(setup_entry.format))
            errs = True
        if (s.schcon.dovex or s.schcon.dovsop) and setup_entry.frswitch:
            s.wlog(1, "CHKSET:  Cannot frequency switch with VEX or VSOP file.")

        if len(setup_entry.channel) < 1:
            s.wlog(1, "CHKSET: Setup file must have at least 1 channel unless "
                   "OBSTYPE=VLA.")
            s.errlog("CHKSET: The error is in {}".format(setup_entry.setname))
            
        overwarn = False
        for ich, channel in enumerate(setup_entry.channel):
            if channel.sidebd not in ("U", "L"):
                s.wlog(1, "CHKSET: Sideband not U or L in {}".format(
                    setup_entry.setname))
                errs = True
            if channel.pol[:3] not in ("RCP", "LCP", "X", "Y"):
                s.wlog(1, "CHKSET: In setup {} polarization of chan {} "
                       "not given or deduced (should be RCP, LCP, X or Y).".\
                       format(setup_entry.setname, ich+1))
                errs = True

            for jch, other in enumerate(setup_entry.channel[ich+1:], ich+1):
                if (channel.pol != other.pol) and \
                   (channel.ifchan == other.ifchan):
                    s.wlog(1, "CHKSET:  In setup {} IF channel {} is "
                           "assigned to {} in chan {} and to {} in chan {} "
                           "Not possible.".format(setup_entry.setname, 
                                                  channel.ifchan, 
                                                  channel.pol, 
                                                  ich+1, 
                                                  channel.pol, 
                                                  jch+1))
                    errs = True
            
            if s.schn1.vlbitp:
                if channel.bbfilt > 0.5001 * setup_entry.samprate:
                    s.wlog(1, "CHKSET: Bandwidth more than half of sample rate")
                    errs = True
                if (channel.bbfilt < 0.4999 * setup_entry.samprate) \
                   and sampwarn:
                    s.wlog(0, "CHKSET note: Oversampling specified.")
                    sampwarn = False

            if ich > 0:
                freq_i = freq_range(channel)
                for jch, other in enumerate(setup_entry.channel[:ich-1]):
                    if channel.pol == other.pol:
                        freq_j = freq_range(other)
                        if (freq_j[0] <= freq_i[0] < freq_j[1]) or \
                           (freq_j[0] < freq_i[1] <= freq_j[1]):
                            overwarn = True
        # end of channel loop
        if overwarn:
            s.wlog(1, "CHKSET: Setup file: {} Has overlapping channels.  "
                   "Intended?".format(setup_entry.setname))

        if station_entry.dar.startswith("RDBE"):
            errs = s.chkrdbe(ks, errs)
        if station_entry.dar.startswith("DBBC"):
            errs = chkdbbc(ks, setup_entry, station_entry) or errs
        if station_entry.dar == "eMERL":
            errs = check_emerlin(setup_entry, station_entry) or errs
        if station_entry.dar == "WIDAR":
            errs = s.chkwidar(ks, errs)
        if station_entry.dar in ("VLBA", "VLBAG", "VLBA4"):
            errs = s.chkvdar(ks, station_entry.nbbc, errs)
        if station_entry.dar == "VLBA4":
            errs = s.chkv4dar(ks, errs)
        if station_entry.dar == "VLBAG":
            errs = s.chkgdar(ks, errs)
        if station_entry.dar == "MKIV":
            errs = s.chk4dar(ks, station_entry.nbbc, errs)
        if station_entry.dar == "CDAS":
            errs = s.chkcdas(ks, errs)
        
        if s.schn1.vlbitp:
            if station_entry.usedisk:
                errs = s.chkdisk(ks, errs)

            if station_entry.recorder in ("VLBA", "MKIV", "VLBA4") and \
               (setup_entry.format != "NONE"):
                errs = s.chkspd(ks, errs)

            if len(np.unique(setup_entry.bits)) > 1:
                s.wlog(1, "CHKSET: All channels must use the same number of "
                       "bits.")
                errs = True

    # end of if not vlaonly
    
    if warn2cm and (15100 < setup_entry.channel[0].freqref < 15500) and \
       setup_entry.setsta[0].startswith("VLBA") and (setup_entry.totbps < 1000):
        s.wlog(1, " ")
        s.wlog(1, "CHKSET:  See sched.runlog for information on 2cm "
               "frequencies.")
        s.wrtmsg(0, "CHKSET", "warn2cm")
        warn2cm = False

    if setup_entry.setsta[0].startswith("VLBA"):
        errs = s.chkvlba(ks, errs)
    if setup_entry.setsta[0].startswith("VLA"):
        errs = s.chkvla(ks, errs)

    if errs:
        s.wlog(1, "CHKSET:  Freq groups used or checked:")
        frequencies = FrequencyCatalog().read()
        for ich, channel in enumerate(setup_entry.channel):
            if channel.ifreqnum > 0:
                s.wlog(1, " Channel: {}  Frequency Group: {}".format(
                    ich+1, frequencies[channel.ifreqnum-1].frname))
        
        s.errset(ks)



import warnings
import logging
from contextlib import contextmanager

# using toolmanager prints warning messages since it's considered experimental
# since version 1.5 (at least up till version 3.3)
@contextmanager
def shut_up_mpl():
    warnings.filterwarnings("ignore")
    tool_manager_logger = logging.getLogger("matplotlib.backend_managers")
    filter_all = lambda record: False
    tool_manager_logger.addFilter(filter_all)
    yield
    tool_manager_logger.removeFilter(filter_all)
    warnings.resetwarnings()




from ..catalog import SourceCatalog
from .. import util, key
from . import parameter

import schedlib as s

import numpy as np

import re
import math
import itertools

eqwarn = True
lastfile = ""
malias = s.schcsc.source.shape[0]
record_defaults = {}
state_defaults = {}

# wrapped version of function below (srread) that opens the filename
# before reading from the input
def srread_open(filename, stdin, select, thiscat):
    if select and s.schsou.msrc > 0:
        gotall = s.srcflg()
    else:
        gotall = False
    
    if ((filename.upper() == "NONE") or (gotall and not s.schcon.plot)):
        return

    try:
        f = open(filename, "r")
    except Exception as e:
        s.putout(str(e))
        s.error("Source catalog: {} not opened".format(filename))
    with f:
        return srread(key.KeyfileLister(f), stdin, select, thiscat)

def srread(input_iterator, stdin, select, thiscat):
    global eqwarn, lastfile, record_defaults, state_defaults

    if s.schcon.debug:
        s.wlog(0, "SRREAD starting")

    infile = "Program_input" if input_iterator.input_ is stdin else \
             input_iterator.input_.name

    if infile != lastfile:
        lastfile = infile
        state_defaults = {
            "version": ["Not given", util.noop],
            "epoch":   [0.,          util.noop],
            "equinox": ["",          util.upper],
            "vdef":    ["R",         util.upper],
            "vref":    ["L",         util.upper],
        }
        record_defaults = {
            "source":   [["NONAME"] + [""] * (malias - 1), 
                                    util.foreach(util.strip_upper)],
            "dec":      [-324000.,  util.multiply_by(parameter.radds)],
            "ra":       [0.,        util.multiply_by(parameter.radhs)],
            "raerr":    [0.,        util.noop],
            "decerr":   [0.,        util.noop],
            "remarks":  ["",        util.noop],
            "calcode":  [" ",       util.upper],
            "velocity": [[-1e9],    util.extend_to(s.schsou.vlsr.shape[0])],
            "dra":      [0.,        util.noop],
            "ddec":     [0.,        util.noop],
            "pmepoch":  [[0.,] * 4, util.noop],
            "pmra":     [0.,        util.multiply_by(1/1000.)],
            "pmdec":    [0.,        util.multiply_by(1/1000.)],
            "paralax":  [0.,        util.noop],
            "endcat":   [None,      util.noop],
            "parallax": [None,      util.noop],
            "flux":     [None,      util.noop], # value ignored
            "fluxref":  [None,      util.noop], # value ignored
        }

    input_iterator.set_defaults(record_defaults, state_defaults)
    
    attribute_to_key = {
        "racat": "ra",
        "deccat": "dec",
        "remark": "remarks",
        "vlsr": "velocity",
        "velref": "vref",
        "veldef": "vdef",
        "epocht": "epoch",
    }
    # update with identity mapping elements
    attribute_to_key.update(
        {key: key for key in itertools.chain(state_defaults.keys(),
                                             record_defaults.keys())
         if (key not in attribute_to_key.values()) and 
         (key not in {"version", "endcat", "parallax", "fluxref"})})
        

    s.wlog(0, "SRREAD:  Reading source catalog:  {}".format(infile[-(100-24):]))
    
    catalog = SourceCatalog()
    index = s.schsou.msrc
    
    for record in input_iterator:
        values, present = util.merge_record_with_defaults(
            record, record_defaults, state_defaults)
        
        if "endcat" in present:
            break

        if "source" not in present:
            s.wlog(1, " WARNING:  An unnamed source was found in the source "
                   "catalog:")
            s.wlog(1, "           {}".format(infile))
            s.wlog(1, "           There may be an extra '/' somewhere.")
            continue

        for source in values["source"]:
            if set(source) & set(';:=&*$ "'):
                s.wlog(1, "SRREAD: The source name {} contains an illegal "
                       "character for Vex.".format(source))
                s.wlog(1, "        The source is in catalog: {}".format(infile))
                s.wlog(1, '        The illegal characters are (tab) '
                       '(new line) ; : = & * $ " and (space).')
                s.errlog("Change the source name and rerun. ")
                
        srused = False
        keepit = False
        if select:
            # check if any of the source name aliases is in the schedule and
            # hasn't been found
            encoded_alias = [util.resize_string(
                source, s.schc1.srcname.itemsize, "source").encode()
                             for source in values["source"] if source != ""]
            if np.in1d(encoded_alias, 
                       np.where(s.schn1.srcatn[:s.schn1.nsrc] == 0, 
                                s.schc1.srcname[:s.schn1.nsrc], 0)).any():
                keepit = True
                srused = True
        else:
            keepit = True

        # backward compatibility for equinox/epoch mix-up
        valid_equinox = ("J2000", "B1950", "DATE")
        epoch = values["epoch"]
        epeq = (isinstance(epoch, str) and (epoch.upper() in valid_equinox))
        if values["equinox"] not in valid_equinox:
            if epeq:
                if eqwarn:
                    s.wlog(1, "SRREAD:  Please convert source catalog from use "
                           "of EPOCH to EQUINOX.")
                    eqwarn = False
                values["equinox"] = epoch.upper()
                values["epoch"] = 0.
            else:
                s.wlog(1, "SRREAD:  Source catalog contains invalid equinox: "
                       "{}".format(values["equinox"]))
                s.wlog(1, "         EPOCH also could not be interpreted as a "
                       "valid equinox.")
                s.errlog(" EQUINOX must be {} ".format(", ".join(valid_equinox)))
        else:
            if epeq:
                s.wrtmsg(0, "SRREAD", "eqep")
                s.errlog("SRREAD: See sched.runlog. Convert source lists to use "
                         "EQUINOX.")
            else:
                if isinstance(epoch, str):
                    s.errlog("SRREAD: Source catalog epoch is a string, but not "
                             "a valid equinox")
                if (epoch != 0.) and \
                   ((values["epoch"] < 1900.) or (values["epoch"] > 2100.)):
                    s.wlog(1, "SRREAD: *****   Very odd EPOCH: {} "
                           "for source {}".format(epoch, values["source"][0]))
                    s.wlog(1, "                Please read sched.runlog.")
                    s.wrtmsg(0, "SRREAD", "goofyepoch")

        if keepit:
            if index >= catalog.maxsource:
                s.errlog("'Too many sources for SCHED.  Max ={}".format(
                    catalog.maxsource))
            entry = catalog.entries[index]

            if ("ra" not in present) or ("dec" not in present):
                s.wlog(1, "SRREAD: WARNING: Default RA or Dec used for {}.  "
                       "Was that intended?".format(values["source"][0]))

            entry.set_keyin_values(values, attribute_to_key)

            entry.whichcat = thiscat
            (entry.c1950, entry.c2000, entry.cdata) = (" ", " ", " ")
            if values["equinox"] == "B1950":
                entry.ra1950 = entry.racat
                entry.d1950 = entry.deccat
                entry.c1950 = "*"
            elif values["equinox"] == "J2000":
                entry.ra2000 = entry.racat
                entry.d2000 = entry.deccat
                entry.c2000 = "*"
            elif values["equinox"] == "DATE":
                entry.rap = entry.racat
                entry.decp = entry.deccat
                entry.cdate = "*"
            else:
                s.errlog("SRREAD: Programming error with equinox")
                
            if not re.match("[A-Z0-9 ]", entry.calcode[0]):
                s.errlog("SRREAD: Illegal character {} in CALCODE for {}".format(
                    entry.calcode[0], entry.source[0]))

            if ("dra" in present) or ("ddec" in present):
                entry.pmtime, error = s.sla_cldj(*values["pmepoch"][:3])
                if error not in [0, 3]:
                    s.errlog(" SRREAD: Problem converting planetary motion "
                             "epoch to MJD")
                entry.pmtime += values["pmepoch"][3] * parameter.radhs / \
                                (math.pi * 2)

            # prefer correct spelling over paralax
            if "parallax" in present:
                entry.paralax = values["parallax"]
            entry.paralax /= 1000.
            if (set(("pmra", "pmdec", "paralax", "parallax")) & present) and \
               (entry.epocht < 1900. or entry.epocht > 2100.):
                s.wlog(1, "SRREAD: **** Proper motions specified for {}".format(
                    entry.source[0]))
                s.wlog(1, "             But a very unlikely EPOCH given: {}".\
                       format(entry.epocht))
                
            index += 1
        # end of if keepit
        s.srlist(values["source"][0], values["ra"], values["dec"], 
                 values["raerr"], values["decerr"], values["equinox"], 
                 values["calcode"], srused)
            
    # end of record loop
    catalog.write(range(index))
    s.schsou.msrc = index
    if thiscat == "1":
        s.schcsc.srver = util.resize_string(
            state_defaults["version"][0], s.schcsc.srver.itemsize, "version")
    elif thiscat == "3":
        s.schcsc.srver2 = util.resize_string(
            state_defaults["version"][0], s.schcsc.srver2.itemsize, "version")
        



from .catalog import Catalog

import schedlib as s

class PhaseCenterCatalog(Catalog):
    """
    CHARACTER        CTRNAME(MCENT)*12
    CHARACTER        CTRSRCN(MCSRC,MCENT)*12
    INTEGER          CTRSRCI(MCSRC,MCENT)
    INTEGER          NCSRC(MCENT)
    """
    maxphasecenter = s.schsou.ncsrc.shape[0]
    block_items = {
        s.schcsc: [
            "ctrname",
            "ctrsrcn"
        ],
        s.schsou: [
            "ctrsrci",
            "ncsrc"
        ]
    }

    def __init__(self):
        super().__init__(self.maxphasecenter, self.block_items)

    def write(self, indices=None):
        if indices is None:
            entries = self.entries
        else:
            entries = [self.entries[i] for i in indices]
        for entry in entries:
            entry.ncsrc = len(entry.ctrsrcn)
        return super().write(indices)

    def adjust_lengths(self, entries):
        for entry in entries:
            length = entry.ncsrc
            entry.ctrsrcn = entry.ctrsrcn[:length]
            entry.ctrsrci = entry.ctrsrci[:length]

    def scheduled_slice(self):
        return slice(s.schsou.ncent)



from . import parameter

import schedlib as s

def toggle(values, present, key1, key2, default):
    if (key1 in present) and (key2 in present):
        s.errlog("TOGGLE: You can't set both {} and {}!".format(key1, key2))
    elif (key1 not in present) and (key2 not in present):
        return default
    elif key1 in present:
        return values[key1]
    else:
        return not values[key2]




from . import toggle

import schedlib as s

def infdb(values, present, entries, index):
    entry = entries[index]

    if s.schcon.debug and (index < 2):
        s.wlog(0, "INFDB: Starting")

    if ((entry.crdch1 != 0) or (entry.crdsetch[0] != 0)) and \
       (entry.crdnch == 0):
        s.errlog("'INFDB:  If CRDCH1 or CRDSETCH is set, "
                 "CRDNCH must also be set.")
    if (entry.crdch1 != 0) and (entry.crdsetch[0] != 0):
        s.errlog("INFDB:  Do not use both CRDCH1 and CRDSETCH!")
    maxcrd = s.schn2a.crdsetch.shape[0]
    if entry.crdnch > maxcrd:
        s.errlog("INFDB:  CRDNCH set too high.  Maximum is {}.  It is {} "
                 "in input scan {}".format(maxcrd, entry.crdnch, index+1))
    if entry.crdch1 != 0:
        entry.crdsetch = [entry.crdch1+i for i in range(entry.crdnch)]

    default = False
    if index > 0:
        default = entries[index-1].crddop
    entry.crddop = toggle(values, present, "crddop", "crdnodop", default)
    entry.gotcrd = entry.crddop or any(freq > 0 for freq in entry.crdfreq)

    if ((entry.crdnch < 1) or (entry.crdnch > maxcrd)) and \
       entry.gotcrd:
        s.wlog(1, "INFDB: CRDNCH is required and must be between 1 and {}".\
               format(maxcrd))
        s.wlog(1, "       when CRDFREQ or CRDDOP is specified.")
        s.errlog("       CRDNCH is {} on scan {}".format(entry.crdnch, index+1))

    if entry.crddop and (entry.crdsetch[0] < 1):
        s.errlog("INFDB:  If invoking DOPCRD, please also set CRDCH1 or "
                 "CRDSETCH")

    if entry.dopsrc == "":
        entry.dopsrc = entry.scnsrc

    if entry.dopcal:
        s.wlog(1, "INFREQ: Use DOPPLER, not DOPCAL.")
    else:
        default = False
        if index > 0:
            default = entries[index-1].dopcal
        entry.dopcal = toggle(values, present, "doppler", "nodop", default)

    # dopincr's scan index dimension is the first, instead of the last,
    # so it doesn't fit into the catalog logic, write into the common block 
    # manually here
    s.schn2b.dopincr[index, :] = values["dopincr"]

    if (index != 0) and (entry.freq[0] != 0) and \
       (entry.setnum != entries[index-1].setnum) and \
       (entry.freq[0] == entries[index-1].freq[0]):
            s.wlog(1, "INFDB: Setup changed and specified frequency didn't. ")
            s.wlog(1, " Is this really what you wanted? ")
            s.wlog(1, " Use FREQ=0 to get the new setup default.")



from . import parameter
from .. import util

import schedlib as s

import numpy as np

def times(lst, start, stop, day, year):
    # the parameter LST is both a boolean flag (implemented by floats) 
    # and a string for the LST station
    use_lst = True
    if isinstance(lst, (float, int)):
        if lst == 0:
            lst = "VLA"
        else:
            use_lst = False
    elif isinstance(lst, str):
        lst = lst.upper()
    if use_lst:
        indices = np.argwhere(s.schcst.station == util.resize_string(
            lst, s.schcst.station.itemsize, "lst").encode())
        if len(indices) == 0:
            s.errlog("TIMES: LST specified but station not in catalog")

        s.schn1.lststa = indices[0] + 1
        s.schn1.lst = True
    else:
        s.schn1.lst = False

    for index in range(s.schn1.nscans):
        s.schn2a.duronly[index] = 0
        if s.schn2b.dur[index] > 0:
            s.schn2a.duronly[index] += 1
        if start[index] != parameter.unset:
            s.schn2a.duronly[index] += 2
        if stop[index] != parameter.unset:
            s.schn2a.duronly[index] += 4
        s.sattim(index+1, start[index], stop[index], day[index], year[index])

    s.schn1.iatutc = s.sla_dat(s.schn2c.startj[0])



from . import makescn
from ..catalog import StationCatalog, ScanCatalog, SetupFileCatalog, \
    SourceCatalog
from ..util import f2str, bool2str

import schedlib as s

import numpy as np
from astropy.coordinates import SkyCoord

station_catalog = StationCatalog()
scan_catalog = ScanCatalog()
sf_catalog = SetupFileCatalog()
source_catalog = SourceCatalog()
def geochk(j_scan, scan_index, start_time, end_time):
    ok_geo = np.empty(shape=s.schsou.geosrci.shape, dtype=bool)
    use_geo = np.empty(shape=s.schsou.geosrci.shape, dtype=int)

    scans = scan_catalog.direct_access_entries
    stations = station_catalog.used(use_direct_access=True)
    sources = source_catalog.entries

    seg_elevation = np.empty(dtype=float, shape=(len(stations), s.schsou.ngeo))

    use_time = True
    last_scan_index = np.zeros(shape=(StationCatalog.maxsta,))
    approx_time = (start_time + end_time) / 2
    n_reject = 0

    if len(stations) > 20:
        s.wlog(1, "GEOCHK:  Printing only first 20 stations information")
        ms_print = 20
    else:
        ms_print = len(stations)

    s.wlog(1, "             --------------------- ")
    year, day, time_rad = s.timej(approx_time)
    c_time = f2str(s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
    s.wlog(1, "Building geodetic segment centered at {} {} {}".format(
        year, day, c_time))
    s.wlog(1, "   Using GEOPRT={:>4}     GEOTRIES={:>4}    GEOBACK={:>3}     "
           "GEOSREP={:>3}".format(s.schsou.geoprt, s.schsou.geotries, 
                                  s.schsou.geoback, s.schsou.geosrep))
    s.wlog(1, "         GEOSLEW={:>6.2f}  GEOSLOW={:>7.1f}  GELOWEL={:>6.2f}  "
           "GEOHIEL={:>6.2f}".format(s.schsou.geoslew, s.schsou.geoslow, 
                                     s.schsou.geolowel, s.schsou.geohiel))
    s.wlog(1, "         See sched.runlog for details of the build process.")

    if s.schsou.geoprt >= 0:
        s.wlog(0, "       Note in fit, SecZ < 4 treated as 4 to avoid favoring "
               "extreme low elevations.")
        s.wlog(0, "Elevations at center for sources considered are: ")
        s.wlog(0, "                  Prio " + "   ".join(
            s.stcode for s in stations[:20]))

    elevation_tolerance = 0
    
    sun_coord = SkyCoord(*s.sunpos(approx_time), unit="rad")
    t_freq = sf_catalog.entries[scans[j_scan - 1].setnum - 1].sffreq[0] / 1000
    required_separation = 60 * (t_freq ** -0.6)

    for geo_index in range(s.schsou.ngeo):
        l_scan = scan_index + 1
        n_good = makescn(last_scan_index, l_scan, j_scan, 
                         s.schsou.geosrci[geo_index],
                         f2str(s.schcsc.geosrc[geo_index]), approx_time, 
                         scans[j_scan - 1].opminel - elevation_tolerance, 
                         use_time)

        for i, station in enumerate(stations):
            seg_elevation[i, geo_index] = (station.el1[l_scan - 1] + 
                                           station.el2[l_scan - 1]) / 2

        ok_geo[geo_index] = (n_good >= scans[j_scan - 1].opmian)

        source = sources[s.schsou.geosrci[geo_index] - 1]
        source_coord = SkyCoord(source.rap, source.decp, unit="rad")
        source_separation = sun_coord.separation(source_coord).deg
        
        ok_geo[geo_index] = ok_geo[geo_index] and \
                            (source_separation > required_separation)

        if not ok_geo[geo_index]:
            n_reject += 1
            use_geo[geo_index] = 9

        if ok_geo[geo_index]:
            use_geo[geo_index] = 5
            for i, station in enumerate(stations):
                if station.stascn[l_scan - 1]:
                    if (seg_elevation[i, geo_index] <= s.schsou.geolowel + 10):
                        use_geo[geo_index] = 4
            for i, station in enumerate(stations):
                if station.stascn[l_scan - 1]:
                    if (seg_elevation[i, geo_index] <= s.schsou.geolowel):
                        use_geo[geo_index] = 3
            
            n_low = 0
            n_high = 0
            for i, station in enumerate(stations):
                if station.stascn[l_scan - 1] and \
                   (seg_elevation[i, geo_index] > 
                    (scans[j_scan - 1].opminel - elevation_tolerance)) and \
                   (seg_elevation[i, geo_index] <= s.schsou.geolowel):
                    n_low += 1
                if station.stascn[l_scan - 1] and \
                   (seg_elevation[i, geo_index] >= s.schsou.geohiel):
                    n_high += 1

            if ((n_low >= 1) and (n_high >= 3)) or (n_low >= 3):
                use_geo[geo_index] = 2
            if (n_low >= 2) and (n_high >= 2):
                use_geo[geo_index] = 1

        if s.schsou.geoprt >= 0:
            msg = "{:>5d} {:<8} {:>5d}".format(
                geo_index + 1, f2str(s.schcsc.geosrc[geo_index]), 
                use_geo[geo_index])
            msg += "".join("{:5.0f}".format(el) 
                           if el > (scans[l_scan - 1].opminel - 
                                    elevation_tolerance)
                           else "   --"
                           for el in seg_elevation[:, geo_index])
            if source_separation <= required_separation:
                msg += "   Too near sun: {:5.0f} deg.".format(source_separation)
            s.wlog(0, msg)
    
    if n_reject == s.schsou.ngeo:
        s.errlog("GEOCHK:  None of the sources specified for a geodetic "
                 "segment are up at OPMINANT antennas.")

    return ok_geo, use_geo, seg_elevation
